# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import heapq
from unittest import TestCase

from grpc import ServicerContext, StatusCode
from mock import Mock, MagicMock, patch, PropertyMock
from pyqrllib.pyqrllib import str2bin, hstr2bin, bin2hstr

from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.PaginatedBitfield import PaginatedBitfield
from qrl.core.Block import Block
from qrl.core.TransactionMetadata import TransactionMetadata
from qrl.core.ChainManager import ChainManager
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.State import State
from qrl.core.TransactionInfo import TransactionInfo
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.misc import logger
from qrl.core.node import SyncState, POW
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.qrlnode import QRLNode
from qrl.crypto.misc import sha256
from qrl.generated import qrl_pb2
from qrl.services.PublicAPIService import PublicAPIService
from tests.blockchain.MockedBlockchain import MockedBlockchain
from tests.misc.helper import get_alice_xmss, get_bob_xmss, replacement_getTime, set_qrl_dir


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestPublicAPI(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPublicAPI, self).__init__(*args, **kwargs)

    def test_getNodeState(self):
        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.sync_state = SyncState()
        p2p_factory.num_connections = 23
        p2p_factory.pow = Mock()

        chain_manager = Mock(spec=ChainManager)
        chain_manager.height = 0
        chain_manager.last_block = Block()

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow

        service = PublicAPIService(qrlnode)
        node_state = service.GetNodeState(request=qrl_pb2.GetNodeStateReq, context=None)

        # self.assertEqual(__version__, node_state.info.version)  # FIXME
        self.assertEqual(qrl_pb2.NodeInfo.UNSYNCED, node_state.info.state)
        self.assertEqual(23, node_state.info.num_connections)
        # self.assertEqual("testnet", node_state.info.network_id)  # FIXME

    def test_getKnownPeers(self):
        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.sync_state = SyncState()
        p2p_factory.num_connections = 23
        p2p_factory.pow = Mock()

        chain_manager = Mock(spec=ChainManager)
        chain_manager.height = 0
        chain_manager.last_block = Block()

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow
        qrlnode.peer_manager = Mock()
        qrlnode.peer_manager.known_peer_addresses = ['127.0.0.1', '192.168.1.1']

        service = PublicAPIService(qrlnode)
        response = service.GetKnownPeers(request=qrl_pb2.GetKnownPeersReq, context=None)

        self.assertEqual(2, len(response.known_peers))
        self.assertEqual('127.0.0.1', response.known_peers[0].ip)
        self.assertEqual('192.168.1.1', response.known_peers[1].ip)

        logger.info(response)

    def test_getStats(self):
        db_state = Mock(spec=State)
        db_state.total_coin_supply = 1000
        db_state.get_measurement = MagicMock(return_value=60)
        db_state.get_block_by_number = MagicMock(return_value=None)

        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.sync_state = SyncState()
        p2p_factory.num_connections = 23
        p2p_factory.pow = Mock()

        chain_manager = ChainManager(db_state)
        chain_manager._last_block = GenesisBlock()

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow

        service = PublicAPIService(qrlnode)
        request = qrl_pb2.GetStatsReq()
        stats = service.GetStats(request=request, context=None)

        # self.assertEqual(__version__, stats.node_info.version)  # FIXME

        self.assertEqual(qrl_pb2.NodeInfo.UNSYNCED, stats.node_info.state)
        self.assertEqual(23, stats.node_info.num_connections)
        # self.assertEqual("testnet", stats.node_info.network_id)  # FIXME

        self.assertEqual(0, stats.epoch)
        self.assertEqual(0, stats.uptime_network)

        self.assertEqual(65000000000000000, stats.block_last_reward)
        self.assertEqual(0, stats.block_time_mean)
        self.assertEqual(0, stats.block_time_sd)

        self.assertEqual(105000000, stats.coins_total_supply)
        self.assertEqual(1000, stats.coins_emitted)

        self.assertEqual(0, len(stats.block_timeseries))

    def test_getStats_timeseries(self):
        with MockedBlockchain.create(10) as mock_blockchain:
            service = PublicAPIService(mock_blockchain.qrlnode)
            request = qrl_pb2.GetStatsReq(include_timeseries=1)
            stats = service.GetStats(request=request, context=None)

            print(stats.block_timeseries)

            self.assertEqual(11, len(stats.block_timeseries))
            self.assertEqual(61, stats.block_time_mean)
            self.assertEqual(1, stats.block_time_sd)

    def test_getAddressState(self):
        with set_qrl_dir('no_data'):
            db_state = State()
            alice_xmss = get_alice_xmss()
            optimized_address_state = OptimizedAddressState.create(address=alice_xmss.address,
                                                                   nonce=25,
                                                                   balance=10,
                                                                   ots_bitfield_used_page=0,
                                                                   transaction_hash_count=0,
                                                                   tokens_count=0,
                                                                   lattice_pk_count=0,
                                                                   slaves_count=0,
                                                                   multi_sig_address_count=0)
            AddressState.put_address_state(db_state, optimized_address_state)

            p2p_factory = Mock(spec=P2PFactory)
            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetAddressStateReq()
            service.GetAddressState(request=request, context=context)
            context.set_code.assert_called()
            context.set_details.assert_called()

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetAddressStateReq()
            request.address = get_alice_xmss().address
            response = service.GetAddressState(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(alice_xmss.address, response.state.address)
            self.assertEqual(25, response.state.nonce)
            self.assertEqual(10, response.state.balance)
            self.assertEqual([b'\x00'] * config.dev.ots_bitfield_size, response.state.ots_bitfield)
            self.assertEqual([], response.state.transaction_hashes)

    def test_getObject(self):
        SOME_ODD_HASH = sha256(b'this should not be found')

        with set_qrl_dir('no_data'):
            db_state = State()

            p2p_factory = Mock(spec=P2PFactory)
            p2p_factory.pow = Mock(spec=POW)

            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._pow = p2p_factory.pow
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            # First try an empty request
            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetObjectReq()
            response = service.GetObject(request=request, context=context)
            context.set_code.assert_not_called()
            context.set_details.assert_not_called()
            self.assertFalse(response.found)

            # Some odd address
            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetObjectReq()
            request.query = SOME_ODD_HASH
            response = service.GetObject(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertFalse(response.found)

            # Find an address
            bob_xmss = get_bob_xmss()
            addr1_state = OptimizedAddressState.create(address=bob_xmss.address,
                                                       nonce=25,
                                                       balance=10,
                                                       ots_bitfield_used_page=0,
                                                       transaction_hash_count=0,
                                                       tokens_count=0,
                                                       lattice_pk_count=0,
                                                       slaves_count=0,
                                                       multi_sig_address_count=0)

            AddressState.put_address_state(db_state, addr1_state, None)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetObjectReq()
            request.query = bob_xmss.address
            response = service.GetObject(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertTrue(response.found)
            self.assertIsNotNone(response.address_state)

            self.assertEqual(bob_xmss.address, response.address_state.address)
            self.assertEqual(25, response.address_state.nonce)
            self.assertEqual(10, response.address_state.balance)

            # Find a transaction
            alice_xmss = get_alice_xmss()
            db_state.address_used = MagicMock(return_value=False)
            tx1 = TransferTransaction.create(
                addrs_to=[bob_xmss.address],
                amounts=[125],
                message_data=None,
                fee=19,
                xmss_pk=bob_xmss.pk,
                master_addr=alice_xmss.address)

            chain_manager.tx_pool.transaction_pool = [(0, TransactionInfo(tx1, 0))]

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetObjectReq()
            request.query = tx1.txhash
            response = service.GetObject(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertTrue(response.found)
            self.assertIsNotNone(response.transaction)
            self.assertEqual('transfer', response.transaction.tx.WhichOneof('transactionType'))
            self.assertEqual(alice_xmss.address, response.transaction.tx.master_addr)
            self.assertEqual(bob_xmss.pk, response.transaction.tx.public_key)
            self.assertEqual(tx1.txhash, response.transaction.tx.transaction_hash)
            self.assertEqual(b'', response.transaction.tx.signature)

            self.assertEqual(bob_xmss.address, response.transaction.tx.transfer.addrs_to[0])
            self.assertEqual(125, response.transaction.tx.transfer.amounts[0])
            self.assertEqual(19, response.transaction.tx.fee)

            alice_xmss = get_alice_xmss()
            # Find a block
            block = Block.create(dev_config=config.dev,
                                 block_number=1,
                                 prev_headerhash=sha256(b'reveal'),
                                 prev_timestamp=10,
                                 transactions=[],
                                 miner_address=alice_xmss.address,
                                 seed_height=0,
                                 seed_hash=None)

            Block.put_block(db_state, block, None)
            Block.put_block_number_mapping(db_state,
                                           block.block_number,
                                           qrl_pb2.BlockNumberMapping(headerhash=block.headerhash),
                                           None)
            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetObjectReq()
            request.query = bytes(str2bin('1'))
            response = service.GetObject(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertTrue(response.found)
            self.assertIsNotNone(response.block_extended)
            self.assertEqual(1, response.block_extended.header.block_number)

    def test_getLatestData(self):
        with set_qrl_dir('no_data'):
            db_state = State()
            chain_manager = ChainManager(db_state)
            blocks = []
            txs = []
            alice_xmss = get_alice_xmss()
            bob_xmss = get_bob_xmss()
            for i in range(0, 4):
                for j in range(1, 3):
                    txs.append(TransferTransaction.create(addrs_to=[bob_xmss.address],
                                                          amounts=[i * 100 + j],
                                                          message_data=None,
                                                          fee=j,
                                                          xmss_pk=alice_xmss.pk))

                blocks.append(Block.create(dev_config=config.dev,
                                           block_number=i,
                                           prev_headerhash=sha256(b'reveal'),
                                           prev_timestamp=10,
                                           transactions=txs,
                                           miner_address=alice_xmss.address,
                                           seed_height=10,
                                           seed_hash=None))
                block = blocks[i]
                Block.put_block(db_state,
                                block,
                                None)
                Block.put_block_number_mapping(db_state,
                                               block.block_number,
                                               qrl_pb2.BlockNumberMapping(headerhash=block.headerhash),
                                               None)
                TransactionMetadata.update_tx_metadata(db_state, block, None)
            chain_manager._last_block = blocks[-1]
            txpool = []
            txs = []
            for j in range(10, 15):
                tx = TransferTransaction.create(addrs_to=[bob_xmss.address],
                                                amounts=[1000 + j],
                                                message_data=None,
                                                fee=j,
                                                xmss_pk=get_alice_xmss().pk)
                txpool.append((tx.fee, TransactionInfo(tx, 0)))
                txs.append(tx)

            p2p_factory = Mock(spec=P2PFactory)
            p2p_factory.pow = Mock(spec=POW)

            chain_manager.tx_pool = Mock()
            chain_manager.tx_pool.transactions = heapq.nlargest(len(txpool), txpool)
            chain_manager.tx_pool.transaction_pool = txpool

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode.get_block_from_index = MagicMock(return_value=None)

            qrlnode._p2pfactory = p2p_factory
            qrlnode._pow = p2p_factory.pow

            service = PublicAPIService(qrlnode)
            context = Mock(spec=ServicerContext)

            request = qrl_pb2.GetLatestDataReq(filter=qrl_pb2.GetLatestDataReq.ALL,
                                               offset=0,
                                               quantity=3)

            response = service.GetLatestData(request=request, context=context)

            context.set_code.assert_not_called()
            context.set_details.assert_not_called()

            # Verify blockheaders
            self.assertEqual(3, len(response.blockheaders))
            self.assertEqual(1, response.blockheaders[0].header.block_number)
            self.assertEqual(2, response.blockheaders[1].header.block_number)
            self.assertEqual(3, response.blockheaders[2].header.block_number)

            # Verify transactions_unconfirmed
            self.assertEqual(3, len(response.transactions_unconfirmed))
            # TODO: Verify expected order
            self.assertEqual(1014, response.transactions_unconfirmed[0].tx.transfer.amounts[0])
            self.assertEqual(1013, response.transactions_unconfirmed[1].tx.transfer.amounts[0])
            self.assertEqual(1012, response.transactions_unconfirmed[2].tx.transfer.amounts[0])

            # Verify transactions
            self.assertEqual(3, len(response.transactions))
            self.assertEqual(2, response.transactions[0].tx.fee)
            self.assertEqual(1, response.transactions[1].tx.fee)
            self.assertEqual(2, response.transactions[2].tx.fee)

            self.assertEqual(302, response.transactions[0].tx.transfer.amounts[0])
            self.assertEqual(301, response.transactions[1].tx.transfer.amounts[0])
            self.assertEqual(202, response.transactions[2].tx.transfer.amounts[0])

    def test_getAddressFromPK(self):
        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.sync_state = SyncState()
        p2p_factory.num_connections = 23
        p2p_factory.pow = Mock()

        chain_manager = Mock(spec=ChainManager)
        chain_manager.height = 0

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow

        service = PublicAPIService(qrlnode)
        pk = hstr2bin('01060057ac9cb6085a8135631dcf018dff46d9c368a0b64d508f512e584199b6800'
                      'f8cfcb672b931398a023680fe0308ed4b6ec75877d684bc2ccf11703e8369f064e7')
        request = qrl_pb2.GetAddressFromPKReq(pk=bytes(pk))
        response = service.GetAddressFromPK(request=request, context=None)
        self.assertEqual('010600b56d161c7de8aa741962e3e49b973b7e53456fa47f2443d69f17c632f29c8b1aab7d2491',
                         bin2hstr(response.address))

    def test_GetTokenTxn_Error(self):
        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.sync_state = SyncState()
        p2p_factory.num_connections = 23
        p2p_factory.pow = Mock()

        chain_manager = Mock(spec=ChainManager)
        chain_manager.height = 0

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow

        service = PublicAPIService(qrlnode)
        request = qrl_pb2.TokenTxnReq()
        context = Mock(spec=ServicerContext)
        context.set_code = MagicMock()

        service.GetTokenTxn(request=request, context=context)
        context.set_code.assert_called_with(StatusCode.INVALID_ARGUMENT)

    @patch('qrl.core.config.DevConfig', new_callable=PropertyMock)
    def test_getMiniTransactionsByAddress(self, mock_dev_config):
        mock_dev_config.data_per_page = 5
        with set_qrl_dir('no_data'):
            db_state = State()

            p2p_factory = Mock(spec=P2PFactory)
            p2p_factory.pow = Mock(spec=POW)

            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._pow = p2p_factory.pow
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            # Find a transaction
            alice_xmss = get_alice_xmss(8)

            context = Mock(spec=ServicerContext)
            bob_xmss = get_bob_xmss(4)
            txs = []
            for i in range(0, 100):
                tx = TransferTransaction.create(addrs_to=[bob_xmss.address],
                                                amounts=[10],
                                                message_data=None,
                                                fee=0,
                                                xmss_pk=alice_xmss.pk)
                tx.sign(alice_xmss)
                txs.append(tx)
            addresses_set = {bob_xmss.address, alice_xmss.address}
            block = Block.create(config.dev, 10, b'', 100000000, txs,
                                 alice_xmss.address, seed_height=0, seed_hash=None)
            state_container = chain_manager.new_state_container(addresses_set,
                                                                5,
                                                                True,
                                                                None)
            for tx in txs:
                chain_manager.update_state_container(tx, state_container)

                address_state = state_container.addresses_state[tx.addr_from]
                state_container.paginated_tx_hash.insert(address_state, tx.txhash)

                address_state = state_container.addresses_state[tx.addrs_to[0]]
                state_container.paginated_tx_hash.insert(address_state, tx.txhash)
            state_container.paginated_tx_hash.put_paginated_data(None)
            OptimizedAddressState.put_optimized_addresses_state(db_state, state_container.addresses_state)

            TransactionMetadata.update_tx_metadata(db_state, block, None)
            request = qrl_pb2.GetMiniTransactionsByAddressReq(address=alice_xmss.address,
                                                              item_per_page=10,
                                                              page_number=1)
            response = service.GetMiniTransactionsByAddress(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(len(response.mini_transactions), 10)
            for i in range(0, 10):
                self.assertEqual(bin2hstr(txs[len(txs) - i - 1].txhash),
                                 response.mini_transactions[i].transaction_hash)
            self.assertEqual(response.balance, 0)

            TransactionMetadata.update_tx_metadata(db_state, block, None)
            request = qrl_pb2.GetMiniTransactionsByAddressReq(address=alice_xmss.address,
                                                              item_per_page=8,
                                                              page_number=3)
            response = service.GetMiniTransactionsByAddress(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(len(response.mini_transactions), 8)
            for i in range(0, 8):
                self.assertEqual(bin2hstr(txs[len(txs) - i - 17].txhash),
                                 response.mini_transactions[i].transaction_hash)
            self.assertEqual(response.balance, 0)

    def test_getTransaction(self):
        db_state = Mock(spec=State)
        db_state.get_tx_metadata = MagicMock(return_value=None)
        db_state.get_block = MagicMock(return_value=None)

        p2p_factory = Mock(spec=P2PFactory)
        p2p_factory.pow = Mock(spec=POW)

        chain_manager = ChainManager(db_state)

        qrlnode = QRLNode(mining_address=b'')
        qrlnode.set_chain_manager(chain_manager)
        qrlnode._p2pfactory = p2p_factory
        qrlnode._pow = p2p_factory.pow
        qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

        service = PublicAPIService(qrlnode)

        # Find an address
        bob_xmss = get_bob_xmss()

        # Find a transaction
        alice_xmss = get_alice_xmss()
        db_state.address_used = MagicMock(return_value=False)
        tx1 = TransferTransaction.create(
            addrs_to=[bob_xmss.address],
            amounts=[125],
            message_data=None,
            fee=19,
            xmss_pk=bob_xmss.pk,
            master_addr=alice_xmss.address)

        chain_manager.tx_pool.transaction_pool = [(0, TransactionInfo(tx1, 0))]

        context = Mock(spec=ServicerContext)
        request = qrl_pb2.GetTransactionReq()
        request.tx_hash = tx1.txhash
        response = service.GetTransaction(request=request, context=context)
        context.set_code.assert_not_called()
        self.assertIsNotNone(response.tx)
        self.assertEqual('transfer', response.tx.WhichOneof('transactionType'))
        self.assertEqual(alice_xmss.address, response.tx.master_addr)
        self.assertEqual(bob_xmss.pk, response.tx.public_key)
        self.assertEqual(tx1.txhash, response.tx.transaction_hash)
        self.assertEqual(b'', response.tx.signature)

        self.assertEqual(bob_xmss.address, response.tx.transfer.addrs_to[0])
        self.assertEqual(125, response.tx.transfer.amounts[0])
        self.assertEqual(19, response.tx.fee)

    def test_getBalance(self):
        with set_qrl_dir('no_data'):
            db_state = State()
            alice_xmss = get_alice_xmss()
            optimized_address_state = OptimizedAddressState.create(address=alice_xmss.address,
                                                                   nonce=25,
                                                                   balance=10,
                                                                   ots_bitfield_used_page=0,
                                                                   transaction_hash_count=0,
                                                                   tokens_count=0,
                                                                   lattice_pk_count=0,
                                                                   slaves_count=0,
                                                                   multi_sig_address_count=0)
            addresses_state = {optimized_address_state.address: optimized_address_state}
            OptimizedAddressState.put_optimized_addresses_state(db_state, addresses_state)

            p2p_factory = Mock(spec=P2PFactory)
            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetAddressStateReq()
            service.GetAddressState(request=request, context=context)
            context.set_code.assert_called()
            context.set_details.assert_called()

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetBalanceReq()
            request.address = get_alice_xmss().address
            response = service.GetBalance(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(10, response.balance)

    def test_getTotalBalance(self):
        with set_qrl_dir('no_data'):
            db_state = State()

            xmss1 = get_alice_xmss()
            xmss2 = get_alice_xmss(4)
            xmss3 = get_bob_xmss(4)
            address_state1 = OptimizedAddressState.create(address=xmss1.address,
                                                          nonce=25,
                                                          balance=1000,
                                                          ots_bitfield_used_page=0,
                                                          transaction_hash_count=0,
                                                          tokens_count=0,
                                                          lattice_pk_count=0,
                                                          slaves_count=0,
                                                          multi_sig_address_count=0)
            address_state2 = OptimizedAddressState.create(address=xmss2.address,
                                                          nonce=25,
                                                          balance=2000,
                                                          ots_bitfield_used_page=0,
                                                          transaction_hash_count=0,
                                                          tokens_count=0,
                                                          lattice_pk_count=0,
                                                          slaves_count=0,
                                                          multi_sig_address_count=0)
            address_state3 = OptimizedAddressState.create(address=xmss3.address,
                                                          nonce=25,
                                                          balance=3000,
                                                          ots_bitfield_used_page=0,
                                                          transaction_hash_count=0,
                                                          tokens_count=0,
                                                          lattice_pk_count=0,
                                                          slaves_count=0,
                                                          multi_sig_address_count=0)
            AddressState.put_address_state(db_state, address_state1)
            AddressState.put_address_state(db_state, address_state2)
            AddressState.put_address_state(db_state, address_state3)

            p2p_factory = Mock(spec=P2PFactory)
            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetAddressStateReq()
            service.GetAddressState(request=request, context=context)
            context.set_code.assert_called()
            context.set_details.assert_called()

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetTotalBalanceReq()
            request.addresses.extend([xmss1.address, xmss2.address, xmss3.address])
            response = service.GetTotalBalance(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(6000, response.balance)

    def test_getOTS(self):
        with set_qrl_dir('no_data'):
            db_state = State()
            alice_xmss = get_alice_xmss()
            optimized_address_state = OptimizedAddressState.create(address=alice_xmss.address,
                                                                   nonce=25,
                                                                   balance=10,
                                                                   ots_bitfield_used_page=0,
                                                                   transaction_hash_count=0,
                                                                   tokens_count=0,
                                                                   lattice_pk_count=0,
                                                                   slaves_count=0,
                                                                   multi_sig_address_count=0)
            addresses_state = {optimized_address_state.address: optimized_address_state}
            AddressState.put_addresses_state(db_state, addresses_state)

            p2p_factory = Mock(spec=P2PFactory)
            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetOTSReq(address=alice_xmss.address)
            response = service.GetOTS(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(0, response.next_unused_ots_index)

            paginated_bitfield = PaginatedBitfield(True, db_state._db)
            paginated_bitfield.set_ots_key(addresses_state, optimized_address_state.address, 0)
            paginated_bitfield.put_addresses_bitfield(None)

            response = service.GetOTS(request=request, context=context)
            context.set_code.assert_not_called()

            self.assertEqual(1, response.next_unused_ots_index)

    def test_getHeight(self):
        with set_qrl_dir('no_data'):
            db_state = State()
            alice_xmss = get_alice_xmss()
            optimized_address_state = OptimizedAddressState.create(address=alice_xmss.address,
                                                                   nonce=25,
                                                                   balance=10,
                                                                   ots_bitfield_used_page=0,
                                                                   transaction_hash_count=0,
                                                                   tokens_count=0,
                                                                   lattice_pk_count=0,
                                                                   slaves_count=0,
                                                                   multi_sig_address_count=0)
            addresses_state = {optimized_address_state.address: optimized_address_state}
            AddressState.put_addresses_state(db_state, addresses_state)

            p2p_factory = Mock(spec=P2PFactory)
            chain_manager = ChainManager(db_state)
            chain_manager._last_block = Mock()
            chain_manager._last_block.block_number = 100

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)
            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetHeightReq()
            response = service.GetHeight(request=request, context=context)
            self.assertEqual(response.height, 100)

    def test_getBlock(self):
        with set_qrl_dir('no_data'):
            db_state = State()

            p2p_factory = Mock(spec=P2PFactory)
            p2p_factory.pow = Mock(spec=POW)

            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._pow = p2p_factory.pow
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            alice_xmss = get_alice_xmss()
            b = Block.create(dev_config=config.dev,
                             block_number=1,
                             prev_headerhash=sha256(b'reveal'),
                             prev_timestamp=10,
                             transactions=[],
                             miner_address=alice_xmss.address,
                             seed_height=0,
                             seed_hash=None)
            Block.put_block(db_state, b, None)
            db_state.get_block = MagicMock(return_value=b)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetBlockReq(header_hash=b.headerhash)
            response = service.GetBlock(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertEqual(1, response.block.header.block_number)

    def test_getBlockByNumber(self):
        with set_qrl_dir('no_data'):
            db_state = State()

            p2p_factory = Mock(spec=P2PFactory)
            p2p_factory.pow = Mock(spec=POW)

            chain_manager = ChainManager(db_state)

            qrlnode = QRLNode(mining_address=b'')
            qrlnode.set_chain_manager(chain_manager)
            qrlnode._p2pfactory = p2p_factory
            qrlnode._pow = p2p_factory.pow
            qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

            service = PublicAPIService(qrlnode)

            alice_xmss = get_alice_xmss()
            b = Block.create(dev_config=config.dev,
                             block_number=1,
                             prev_headerhash=sha256(b'reveal'),
                             prev_timestamp=10,
                             transactions=[],
                             miner_address=alice_xmss.address,
                             seed_height=0,
                             seed_hash=None)
            Block.put_block(db_state, b, None)
            Block.put_block_number_mapping(db_state,
                                           b.block_number,
                                           qrl_pb2.BlockNumberMapping(headerhash=b.headerhash),
                                           None)
            db_state.get_block_by_number = MagicMock(return_value=b)

            context = Mock(spec=ServicerContext)
            request = qrl_pb2.GetBlockByNumberReq(block_number=b.block_number)
            response = service.GetBlockByNumber(request=request, context=context)
            context.set_code.assert_not_called()
            self.assertEqual(1, response.block.header.block_number)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.LastTransactions import LastTransactions
from qrl.core.txs.Transaction import Transaction
from qrl.generated import qrlstateinfo_pb2


class TransactionMetadata:
    def __init__(self, pbdata=None):
        self._data = pbdata
        if not pbdata:
            self._data = qrlstateinfo_pb2.TransactionMetadata()

    @property
    def transaction(self):
        return self._data.transaction

    @property
    def block_number(self):
        return self._data.block_number

    @property
    def timestamp(self):
        return self._data.timestamp

    @staticmethod
    def create(tx: Transaction, block_number: int, timestamp: int):
        tm = TransactionMetadata()

        tm._data.transaction.MergeFrom(tx.pbdata)
        tm._data.block_number = block_number
        tm._data.timestamp = timestamp

        return tm

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrlstateinfo_pb2.TransactionMetadata()
        pbdata.ParseFromString(bytes(data))
        return TransactionMetadata(pbdata)

    @staticmethod
    def put_tx_metadata(state: State, txn: Transaction, block_number: int, timestamp: int, batch) -> bool:
        try:
            tm = TransactionMetadata.create(tx=txn,
                                            block_number=block_number,
                                            timestamp=timestamp)
            state._db.put_raw(txn.txhash,
                              tm.serialize(),
                              batch)
        except Exception:
            logger.warning("Error writing tx metadata")
            return False

        return True

    @staticmethod
    def get_tx_metadata(state: State, txhash: bytes):
        try:
            tx_metadata = TransactionMetadata.deserialize(state._db.get_raw(txhash))
            data, block_number = tx_metadata.transaction, tx_metadata.block_number
            return Transaction.from_pbdata(data), block_number
        except Exception:
            return None

    @staticmethod
    def rollback_tx_metadata(state: State, block, batch):
        fee_reward = 0
        for protobuf_txn in block.transactions:
            txn = Transaction.from_pbdata(protobuf_txn)
            fee_reward += txn.fee
            TransactionMetadata.remove_tx_metadata(state, txn, batch)

        txn = Transaction.from_pbdata(block.transactions[0])  # Coinbase Transaction
        state._update_total_coin_supply(fee_reward - txn.amount, batch)
        LastTransactions._remove_last_tx(state, block, batch)

    @staticmethod
    def update_tx_metadata(state: State, block, batch) -> bool:
        fee_reward = 0

        for protobuf_txn in block.transactions:
            txn = Transaction.from_pbdata(protobuf_txn)
            fee_reward += txn.fee
            if not TransactionMetadata.put_tx_metadata(state,
                                                       txn,
                                                       block.block_number,
                                                       block.timestamp,
                                                       batch):
                return False

        txn = Transaction.from_pbdata(block.transactions[0])  # Coinbase Transaction
        state._update_total_coin_supply(txn.amount - fee_reward, batch)
        LastTransactions._update_last_tx(state, block, batch)

        return True

    @staticmethod
    def remove_tx_metadata(state: State, txn, batch) -> bool:
        try:
            state._db.delete(txn.txhash, batch)
        except KeyError:
            logger.warning("Error removing tx metadata")
            return False

        return True



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from mock import patch, PropertyMock
from unittest import TestCase

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.PaginatedData import PaginatedData
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.State import State

from tests.misc.helper import set_qrl_dir, get_alice_xmss

logger.initialize_default()


class TestPaginatedData(TestCase):
    def setUp(self):
        self.alice = get_alice_xmss()

    def test_reset_key_value(self):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)

            self.assertEqual(len(p.key_value), 0)

            p.key_value[b'a'] = [10]
            self.assertEqual(len(p.key_value), 1)

            p.reset_key_value()
            self.assertEqual(len(p.key_value), 0)

    def test_get_value(self):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            storage_key = p.generate_key(b'a', 0)
            p.key_value[storage_key] = [10]
            self.assertEqual(p.get_value(b'a', 0), [10])

    def test_insert(self):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)

            p.insert(OptimizedAddressState.get_default(b'a'), b'10')
            self.assertEqual(p.get_value(b'a', 0), [b'10'])

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_put_paginated_data(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)

            for i in range(0, 10):
                p.insert(alice_address_state, b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))
                self.assertEqual(alice_address_state.get_counter_by_name(p.name), i + 1)

            p.put_paginated_data(None)
            self.assertEqual(alice_address_state.get_counter_by_name(p.name), 10)

            for i in range(10, 25):
                p.insert(alice_address_state, b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))
                self.assertEqual(alice_address_state.get_counter_by_name(p.name), i + 1)

            p.put_paginated_data(None)
            self.assertEqual(alice_address_state.get_counter_by_name(p.name), 25)
            self.assertEqual(len(p.key_value), 0)

            pages_data = []
            for i in range(0, (25 // config.dev.data_per_page) + 1):
                data = p.get_paginated_data(self.alice.address,
                                            (i + 1) * config.dev.data_per_page - 1)
                pages_data.append(data)

            self.assertEqual(len(pages_data), 3)

            self.assertEqual(len(pages_data[0]), 10)
            for i in range(0, 10):
                self.assertEqual(pages_data[0][i], b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))

            self.assertEqual(len(pages_data[1]), 10)
            for i in range(10, 20):
                self.assertEqual(pages_data[1][i - 10], b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))

            self.assertEqual(len(pages_data[2]), 5)
            for i in range(20, 25):
                self.assertEqual(pages_data[2][i - 20], b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_revert_paginated_data(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)

            for i in range(0, 25):
                p.insert(alice_address_state, b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))
            p.put_paginated_data(None)

            full_hash = []
            for i in range(0, (25 // config.dev.data_per_page) + 1):
                data = p.get_paginated_data(self.alice.address,
                                            (i + 1) * config.dev.data_per_page - 1)
                full_hash.extend(data)

            for tx_hash in full_hash[24:18:-1]:
                p.remove(alice_address_state, tx_hash)
            p.put_paginated_data(None)

            self.assertEqual(alice_address_state.get_counter_by_name(b'p_tx_hash'),
                             19)

            # page 2 data count must be 0
            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, 2 * config.dev.data_per_page)),
                             0)

            # page 1 data count must be 9
            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, config.dev.data_per_page)),
                             9)

            # page 0 data count must be 10
            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, 0)),
                             10)

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_revert_paginated_data2(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)

            for i in range(0, 25):
                p.insert(alice_address_state, b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False))

            p.put_paginated_data(None)

            full_hash = []
            for i in range(0, (25 // config.dev.data_per_page) + 1):
                data = p.get_paginated_data(self.alice.address,
                                            (i + 1) * config.dev.data_per_page - 1)
                full_hash.extend(data)

            for tx_hash in full_hash[-1::-1]:
                p.remove(alice_address_state, tx_hash)
            p.put_paginated_data(None)

            self.assertEqual(alice_address_state.get_counter_by_name(b'p_tx_hash'),
                             0)

            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, 22)),
                             0)

            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, 12)),
                             0)

            self.assertEqual(len(p.get_paginated_data(alice_address_state.address, 2)),
                             0)

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_get_paginated_data(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)

            total_hashes = 25
            expected_full_hash = []
            for i in range(0, total_hashes):
                tx_hash = b'p_tx_hash_' + i.to_bytes(8, byteorder='big', signed=False)
                p.insert(alice_address_state, tx_hash)
                expected_full_hash.append(tx_hash)

            p.put_paginated_data(None)

            found_full_hash = []
            expected_data_count = [10, 10, 5]
            for i in range(0, (total_hashes // config.dev.data_per_page) + 1):
                data = p.get_paginated_data(self.alice.address,
                                            (i + 1) * config.dev.data_per_page - 1)
                self.assertEqual(len(data), expected_data_count[i])
                found_full_hash.extend(data)

            self.assertEqual(expected_full_hash, found_full_hash)

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_put(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            key = b'test_key'
            value = [b'hello world']
            storage_key = p.generate_key(key, 11)
            p.put(storage_key, value, None)
            found_value = p.get_paginated_data(key, 11)
            self.assertEqual(value, found_value)

    @patch('qrl.core.config.DevConfig.data_per_page', new_callable=PropertyMock, return_value=10)
    def test_remove(self, mock_dev_config):
        with set_qrl_dir('no_data'):
            state = State()
            p = PaginatedData(b'p_tx_hash', True, state._db)
            key = b'test_key'
            value = [b'hello world']
            storage_key = p.generate_key(key, 11)
            p.put(storage_key, value, None)
            found_value = p.get_paginated_data(key, 11)
            self.assertEqual(value, found_value)

            p.delete(storage_key, None)
            found_value = p.get_paginated_data(key, 11)
            self.assertEqual([], found_value)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from typing import Optional

from qrl.core import config
from qrl.core.misc import logger, db
from qrl.generated import qrl_pb2, qrlstateinfo_pb2


class State:
    def __init__(self, my_db=None):
        self._db = my_db
        if not my_db:
            self._db = db.DB()  # generate db object here
        self._tmp_state = None  # Temporary State file which needs to be fetched during migration to new db
        self._state_version = 1  # Change State Version, each time any change made to leveldb structure

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._db is not None:
            if self._db.db is not None:
                del self._db.db
            del self._db
            self._db = None

    @property
    def state_version(self):
        return self._state_version

    @property
    def batch(self):
        return self._db.get_batch()

    @property
    def total_coin_supply(self):
        try:
            return int.from_bytes(self._db.get_raw(b'total_coin_supply'), byteorder='big', signed=False)
        except KeyError:
            return 0

    def get_state_version(self) -> int:
        try:
            version = self._db.get_raw(b'state_version')
            return int(version.decode())
        except KeyError:
            return 0
        except Exception:
            raise Exception("Exception while retrieving version")

    def put_state_version(self):
        try:
            self._db.put_raw(b'state_version', str(self._state_version).encode())
        except Exception:
            raise Exception("Exception while Setting version")

    def is_older_state_version(self):
        current_state_version = self.get_state_version()
        if current_state_version < self._state_version:
            return True

    def is_state_compatible(self) -> bool:
        current_state_version = self.get_state_version()
        if current_state_version > self._state_version:
            logger.warning("You have a state with Version %s", current_state_version)
            logger.warning("This node only supports State Version %s", self._state_version)
            return False
        elif self.is_older_state_version():
            logger.warning("Old State Version Found %s", current_state_version)

        return True

    def get_mainchain_height(self) -> int:
        try:
            return int.from_bytes(self._db.get_raw(b'blockheight'), byteorder='big', signed=False)
        except KeyError:
            pass
        except Exception as e:
            logger.error('get_blockheight Exception %s', e)

        return -1

    def update_mainchain_height(self, height, batch):
        self._db.put_raw(b'blockheight', height.to_bytes(8, byteorder='big', signed=False), batch)

    def get_re_org_limit(self) -> int:
        try:
            return int.from_bytes(self._db.get_raw(b'reorg_limit'), byteorder='big', signed=False)
        except KeyError:
            return 0
        except Exception as e:
            logger.error('get_re_org_limit Exception %s', e)

        return -1

    def update_re_org_limit(self, height, batch):
        reorg_limit = height - config.dev.reorg_limit
        if reorg_limit <= 0:
            return
        current_reorg_limit = self.get_re_org_limit()
        if reorg_limit <= current_reorg_limit:
            return
        self._db.put_raw(b'reorg_limit', reorg_limit.to_bytes(8, byteorder='big', signed=False), batch)

    def get_address_is_used(self, address: bytes) -> bool:
        # FIXME: Probably obsolete
        try:
            return self._db.get_raw(address)
        except KeyError:
            return False
        except Exception as e:
            # FIXME: Review
            logger.error('Exception in address_used')
            logger.exception(e)
            raise

    def write_batch(self, batch, sync=True):
        self._db.write_batch(batch, sync)

    #########################################
    #########################################
    #########################################
    #########################################
    #########################################

    def _update_total_coin_supply(self, balance, batch):
        self._db.put_raw(b'total_coin_supply',
                         (self.total_coin_supply + balance).to_bytes(8, byteorder='big', signed=False),
                         batch)

    def _delete(self, key, batch):
        self._db.delete(key, batch)

    def put_dev_config_state(self, dev_config, batch):
        self._db.put_raw(dev_config.current_state_key, dev_config.SerializeToString(), batch)

    def get_dev_config_state(self, dev_config_state_key: bytes):
        try:
            data = self._db.get_raw(dev_config_state_key)
            pbdata = qrl_pb2.DevConfig()
            pbdata.ParseFromString(bytes(data))
            return pbdata
        except KeyError:
            logger.debug('[get_dev_config_state] Dev Config not found')
        except Exception as e:
            logger.error('[get_dev_config_state] %s', e)

        return None

    def get_dev_config_current_state_key(self):
        try:
            return self._db.get_raw(b'dev_config_current_state_key')
        except KeyError:
            logger.debug('[get_dev_config_current_state_key] Dev Config not found')
        except Exception as e:
            logger.error('[get_dev_config_current_state_key] %s', e)

        return None

    def put_dev_config_current_state_key(self, dev_config_state_key: bytes, batch):
        self._db.put_raw(b'dev_config_current_state_key', dev_config_state_key, batch)

    def put_fork_state(self, fork_state: qrlstateinfo_pb2.ForkState, batch=None):
        self._db.put_raw(b'fork_state', fork_state.SerializeToString(), batch)

    def get_fork_state(self) -> Optional[qrlstateinfo_pb2.ForkState]:
        try:
            data = self._db.get_raw(b'fork_state')
            fork_state = qrlstateinfo_pb2.ForkState()
            fork_state.ParseFromString(bytes(data))
            return fork_state
        except KeyError:
            return None
        except Exception as e:
            logger.error('Exception in get_fork_state')
            logger.exception(e)
            raise

    def delete_fork_state(self, batch=None):
        self._db.delete(b'fork_state', batch)

    @staticmethod
    def generate_token_key(address, token_txhash) -> bytes:
        return b'token_' + address + b'_' + token_txhash

    @staticmethod
    def generate_slave_key(address, slave_pk) -> bytes:
        return b'slave_' + address + b'_' + slave_pk

    def get_slave_pk_access_type(self, address: bytes, slave_pk: bytes) -> qrl_pb2.SlaveMetadata:
        slave_key = self.generate_slave_key(address, slave_pk)
        try:
            slave_metadata = qrl_pb2.SlaveMetadata()
            slave_metadata.ParseFromString(self._db.get_raw(slave_key))
            return slave_metadata
        except KeyError:
            pass
        except Exception as e:
            logger.error('[get_slave_pk_access_type] %s', e)

        return None

    def get_token(self, address: bytes, token_txhash: bytes) -> qrl_pb2.TokenBalance:
        try:
            token_balance = qrl_pb2.TokenBalance()
            token_balance.ParseFromString(self._db.get_raw(self.generate_token_key(address, token_txhash)))
            return token_balance
        except KeyError:
            pass
        except Exception as e:
            logger.error('[get_token] %s', e)

        return None



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from tests.misc.helper import get_slave_xmss
from qrl.core.misc import logger
from qrl.core.BlockMetadata import BlockMetadata
from qrl.core.State import State

from tests.misc.helper import set_qrl_dir, get_alice_xmss

logger.initialize_default()

alice = get_alice_xmss()
slave = get_slave_xmss()


class TestTokenMetadata(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

    def test_put_block_metadata(self):
        block_metadata = BlockMetadata.create()
        block_metadata.update_last_headerhashes([b'test1', b'test2'], b'test3')

        BlockMetadata.put_block_metadata(self.state, b'block_headerhash', block_metadata, None)
        BlockMetadata.put_block_metadata(self.state, b'block_headerhash2', BlockMetadata.create(), None)

        self.assertEqual(BlockMetadata.get_block_metadata(self.state, b'block_headerhash').to_json(),
                         block_metadata.to_json())

        expected_json = b'{\n  "blockDifficulty": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",\n  ' \
                        b'"cumulativeDifficulty": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA="\n}'

        self.assertEqual(BlockMetadata.get_block_metadata(self.state, b'block_headerhash2').to_json(),
                         expected_json)

    def test_get_block_metadata(self):
        self.assertIsNone(BlockMetadata.get_block_metadata(self.state, b'test1'))
        BlockMetadata.put_block_metadata(self.state, b'block_headerhash2', BlockMetadata.create(), None)

        tmp_json = BlockMetadata.get_block_metadata(self.state, b'block_headerhash2').to_json()

        expected_json = b'{\n  "blockDifficulty": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",\n  ' \
                        b'"cumulativeDifficulty": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA="\n}'

        self.assertEqual(tmp_json, expected_json)



from pyqrllib.pyqrllib import bin2hstr

from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction
from qrl.generated.qrl_pb2 import LatticePKMetadata


class LatticeTransaction(Transaction):

    def __init__(self, protobuf_transaction=None):
        super(LatticeTransaction, self).__init__(protobuf_transaction)

    @property
    def pk1(self):  # kyber_pk
        return self._data.latticePK.pk1

    @property
    def pk2(self):  # dilithium_pk
        return self._data.latticePK.pk2

    @property
    def pk3(self):  # ecdsa_pk
        return self._data.latticePK.pk3

    def get_data_bytes(self):
        return self.master_addr + \
               self.fee.to_bytes(8, byteorder='big', signed=False) + \
               self.pk1 + \
               self.pk2 + \
               self.pk3

    @staticmethod
    def create(pk1: bytes, pk2: bytes, pk3: bytes, fee: int, xmss_pk: bytes, master_addr: bytes = None):
        transaction = LatticeTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        transaction._data.fee = fee
        transaction._data.public_key = xmss_pk

        transaction._data.latticePK.pk1 = bytes(pk1)
        transaction._data.latticePK.pk2 = bytes(pk2)
        transaction._data.latticePK.pk3 = bytes(pk3)

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self) -> bool:
        if self.fee < 0:
            logger.info('State validation failed for %s because: Negative send', bin2hstr(self.txhash))
            return False

        return True

    def _validate_extended(self, state_container: StateContainer) -> bool:
        if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
            logger.warning("[LatticeTransaction] Hard Fork Feature not yet activated")
            return False

        dev_config = state_container.current_dev_config
        if len(self.pk1) > dev_config.lattice_pk1_max_length:  # TODO: to fix kyber pk value
            logger.warning('Kyber PK length cannot be more than %s bytes', dev_config.lattice_pk1_max_length)
            logger.warning('Found length %s', len(self.pk1))
            return False

        if len(self.pk2) > dev_config.lattice_pk2_max_length:  # TODO: to fix dilithium pk value
            logger.warning('Dilithium PK length cannot be more than %s bytes', dev_config.lattice_pk2_max_length)
            logger.warning('Found length %s', len(self.pk2))
            return False

        if len(self.pk3) > dev_config.lattice_pk3_max_length:  # TODO: to fix ecdsa pk value
            logger.warning('ECDSA PK length cannot be more than %s bytes', dev_config.lattice_pk3_max_length)
            logger.warning('Found length %s', len(self.pk3))
            return False

        tx_balance = state_container.addresses_state[self.addr_from].balance

        if tx_balance < self.fee:
            logger.info('State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, amount: %s', tx_balance, self.fee)
            return False

        if (self.addr_from, self.pk1, self.pk2, self.pk3) in state_container.lattice_pk.data:
            logger.info('State validation failed for %s because: Lattice PKs already exists for this address',
                        bin2hstr(self.txhash))
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_lattice_pk.insert(address_state, self.txhash)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        state_container.lattice_pk.data[(self.addr_from,
                                         self.pk1, self.pk2, self.pk3)] = LatticePKMetadata(enabled=True)

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_lattice_pk.remove(address_state, self.txhash)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        state_container.lattice_pk.data[(self.addr_from,
                                         self.pk1, self.pk2, self.pk3)] = LatticePKMetadata(enabled=False,
                                                                                            delete=True)

        return self._revert_state_changes_for_PK(state_container)



from unittest import TestCase

from mock import patch, PropertyMock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from tests.misc.helper import get_alice_xmss, get_slave_xmss, get_bob_xmss, set_qrl_dir


@patch('qrl.core.txs.Transaction.logger')
class TestSlaveTransaction(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestSlaveTransaction, self).__init__(*args, **kwargs)
        with set_qrl_dir('no_data'):
            self.state = State()

    def setUp(self):
        self.alice = get_alice_xmss()
        self.slave = get_slave_xmss()
        self.params = {
            "slave_pks": [self.slave.pk],
            "access_types": [0],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }

    def test_create_validate(self, m_logger):
        """Default self.params should result in a valid SlaveTransaction"""
        tx = SlaveTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_or_raise()
        self.assertTrue(result)

    def test_validate_custom(self, m_logger):
        """
        SlaveTransaction._validate_custom() checks for the following things:
        1. if you specify more than 100 slave_pks at once
        2. if len(slave_pks) != len(access_types)
        3. access_types can only be 0, 1
        """
        # Unequal length slave_pks and access_types
        params = self.params.copy()
        params["slave_pks"] = [self.slave.pk]
        params["access_types"] = [0, 1]
        with self.assertRaises(ValueError):
            SlaveTransaction.create(**params)

        # access_type is a weird, undefined number
        params = self.params.copy()
        params["access_types"] = [5]
        with self.assertRaises(ValueError):
            SlaveTransaction.create(**params)

    @patch('qrl.core.txs.Transaction.Transaction.validate_slave', return_value=True)
    def test_validate_extended(self, m_validate_slave, m_logger):
        """
        SlaveTransaction._validate_extended checks for:
        1. valid master/slave
        2. negative fee,
        3. addr_from has enough funds for the fee
        4. addr_from ots_key reuse
        """
        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100
        addresses_state = {
            alice_address_state.address: alice_address_state
        }

        tx = SlaveTransaction.create(**self.params)
        tx.sign(self.alice)

        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        # Invalid master XMSS/slave XMSS relationship
        m_validate_slave.return_value = False
        result = tx.validate_all(state_container)
        self.assertFalse(result)
        m_validate_slave.return_value = True

        # fee = -1
        with patch('qrl.core.txs.SlaveTransaction.SlaveTransaction.fee', new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            result = tx._validate_custom()
            self.assertFalse(result)

        # balance = 0, cannot pay the Transaction fee
        alice_address_state.pbdata.balance = 0
        result = tx._validate_extended(state_container)
        self.assertFalse(result)
        alice_address_state.pbdata.balance = 100

        addresses_state = {
            self.alice.address: alice_address_state
        }
        # addr_from_pk has used this OTS key before
        state_container.paginated_bitfield.set_ots_key(addresses_state, alice_address_state.address, tx.ots_key)
        result = tx.validate_all(state_container)
        self.assertFalse(result)

        bob = get_bob_xmss()
        # Too many slave_pks
        with patch('qrl.core.config', autospec=True) as m_config:
            m_config.dev = config.dev.create(config.dev.prev_state_key, config.dev.current_state_key,
                                             b'', 10, True, True)
            m_config.dev.pbdata.transaction.multi_output_limit = 2
            state_container.current_dev_config = m_config.dev
            params = self.params.copy()
            params["slave_pks"] = [self.alice.pk, bob.pk, self.slave.pk]
            params["access_types"] = [0, 0, 0]

            tx = SlaveTransaction.create(**params)
            self.assertFalse(tx._validate_extended(state_container))

    def test_validate_tx(self, m_logger):
        tx = SlaveTransaction.create(**self.params)
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()



from pyqrllib.pyqrllib import bin2hstr

from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction
from qrl.crypto.misc import sha256


class MultiSigVote(Transaction):
    """
    MultiSigSpend for the transaction of QRL from a multi sig wallet to another wallet.
    """

    def __init__(self, protobuf_transaction=None):
        super(MultiSigVote, self).__init__(protobuf_transaction)

    @property
    def shared_key(self):
        return self._data.multi_sig_vote.shared_key

    @property
    def unvote(self):
        return self._data.multi_sig_vote.unvote

    @property
    def prev_tx_hash(self):
        return self._data.multi_sig_vote.prev_tx_hash

    def set_prev_tx_hash(self, prev_tx_hash: bytes):
        self._data.multi_sig_vote.prev_tx_hash = prev_tx_hash

    def get_data_hash(self):
        tmp_tx_hash = (self.master_addr +
                       self.fee.to_bytes(8, byteorder='big', signed=False) +
                       self.shared_key +
                       self.unvote.to_bytes(1, byteorder='big', signed=False))

        return sha256(tmp_tx_hash)

    @staticmethod
    def create(shared_key: bytes,
               unvote: bool,
               fee: int,
               xmss_pk,
               master_addr: bytes = None):
        multi_sig_vote = MultiSigVote()

        if master_addr:
            multi_sig_vote._data.master_addr = master_addr

        multi_sig_vote._data.public_key = bytes(xmss_pk)

        multi_sig_vote._data.multi_sig_vote.shared_key = shared_key
        multi_sig_vote._data.multi_sig_vote.unvote = unvote

        multi_sig_vote._data.fee = int(fee)

        multi_sig_vote.validate_or_raise(verify_signature=False)

        return multi_sig_vote

    def _validate_custom(self):
        if self.fee < 0:
            logger.warning('MultiSigVote [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)
            return False

        return True

    def _validate_extended(self, state_container: StateContainer):
        if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
            logger.warning("[MultiSigVote] Hard Fork Feature not yet activated")
            return False

        addr_from_state = state_container.addresses_state[self.addr_from]
        vote_stats = state_container.votes_stats[self.shared_key]
        if vote_stats is None:
            logger.warning("[MultiSigVote] Invalid Shared key %s", bin2hstr(self.shared_key))
            return False
        multi_sig_spend_tx = state_container.multi_sig_spend_txs[self.shared_key]
        block_number = state_container.block_number
        if vote_stats.executed:
            logger.warning("[MultiSigVote] Invalid Tx as MultiSigSpend has already been executed")
            return False

        if multi_sig_spend_tx is None:
            logger.warning("MultiSigSpend not found, Shared Key %s", bin2hstr(self.shared_key))
            return False

        if block_number > multi_sig_spend_tx.expiry_block_number:
            logger.warning("[MultiSigVote] Voted for expired Multi Sig Spend Txn")
            logger.warning("Expiry Block Number: %s, Current Block Number: %s",
                           multi_sig_spend_tx.expiry_block_number,
                           block_number)
            return False

        if self.addr_from not in vote_stats.signatories:
            logger.warning("Address not found in signatory list")
            logger.warning("Address %s, Shared Key %s, Multi Sig Address %s",
                           bin2hstr(self.addr_from),
                           bin2hstr(self.shared_key),
                           bin2hstr(vote_stats.multi_sig_address))
            return False

        index = vote_stats.get_address_index(self.addr_from)
        if vote_stats.unvotes[index] == self.unvote:
            logger.warning("[MultiSigVote] Invalid as Vote type already executed")
            logger.warning("Vote type %s", self.unvote)
            return False

        tx_balance = addr_from_state.balance

        if tx_balance < self.fee:
            logger.warning('[MultiSigVote] State validation failed for %s because: Insufficient funds',
                           bin2hstr(self.txhash))
            logger.warning('balance: %s, fee: %s', tx_balance, self.fee)
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)
        vote_stats = state_container.votes_stats[self.shared_key]
        multi_sig_address = vote_stats.multi_sig_address

        weight, found = state_container.addresses_state[multi_sig_address].get_weight_by_signatory(self.addr_from)
        if not found:
            logger.info("[MultiSigVote] Address is not the signatory for the multi sig address")
            return False

        self.set_prev_tx_hash(vote_stats.get_vote_tx_hash_by_signatory_address(self.addr_from))

        if not vote_stats.apply_vote_stats(self, weight, state_container):
            logger.info("[MultiSigVote] Failed to apply vote_stats")
            return False

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        vote_stats = state_container.votes_stats[self.shared_key]
        multi_sig_address = vote_stats.multi_sig_address

        weight, found = state_container.addresses_state[multi_sig_address].get_weight_by_signatory(self.addr_from)
        if not found:
            logger.info("[MultiSigVote] Address is not the signatory for the multi sig address")
            return False

        if not vote_stats.revert_vote_stats(self, weight, state_container):
            logger.info("[MultiSigVote] Failed to revert vote_stats")
            return False

        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        return self._revert_state_changes_for_PK(state_container)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from mock import PropertyMock, patch
from unittest import TestCase
from math import ceil

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.PaginatedBitfield import PaginatedBitfield
from qrl.core.AddressState import AddressState
from tests.misc.helper import get_alice_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()

alice = get_alice_xmss()
slave = get_slave_xmss()


class TestOptimizedAddressState(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

    # TODO: Move this test to Optimized Address State
    def test_get_optimized_address_state(self):
        alice_xmss = get_alice_xmss()

        alice_address = alice_xmss.address
        address_state = OptimizedAddressState.get_optimized_address_state(self.state, alice_address)
        self.assertTrue(isinstance(address_state.address, bytes))

        alice_address = bytearray(alice_xmss.address)
        with self.assertRaises(TypeError):
            OptimizedAddressState.get_optimized_address_state(self.state, alice_address)

        alice_address = alice_xmss.address
        address_state = OptimizedAddressState.get_optimized_address_state(self.state, alice_address)
        addresses_state = {
            alice_address: address_state
        }
        self.assertTrue(isinstance(address_state.address, bytes))
        AddressState.put_addresses_state(self.state, addresses_state)

        address_state = OptimizedAddressState.get_optimized_address_state(self.state, alice_address)
        self.assertTrue(isinstance(address_state.address, bytes))

    def test_get_optimized_address_state2(self):
        alice_xmss = get_alice_xmss()

        alice_address = alice_xmss.address
        address_state = OptimizedAddressState.get_optimized_address_state(self.state, alice_address)
        addresses_state = {
            alice_address: address_state
        }
        self.assertTrue(isinstance(address_state.address, bytes))
        OptimizedAddressState.put_optimized_addresses_state(self.state, addresses_state)
        address_state = OptimizedAddressState.get_optimized_address_state(self.state, alice_address)
        self.assertTrue(isinstance(address_state.address, bytes))

    def test_update_used_page_in_address_state(self):
        alice_xmss = get_alice_xmss(4)
        address = alice_xmss.address
        address_state = OptimizedAddressState.get_default(address)
        addresses_state = {address: address_state}
        paginated_bitfield = PaginatedBitfield(True, self.state._db)
        key = paginated_bitfield.generate_bitfield_key(address, 1)
        paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
        self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(0, 16):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            if i != 15:
                self.assertEqual(address_state.ots_bitfield_used_page, 0)
            else:
                self.assertEqual(address_state.ots_bitfield_used_page, 1)

        for i in range(0, 16):
            ots_bitfield = paginated_bitfield.key_value[key]
            ots_key_index = i % config.dev.ots_tracking_per_page
            offset = ots_key_index >> 3
            relative = ots_key_index % 8
            bitfield = bytearray(ots_bitfield[offset])
            self.assertEqual(bytes([bitfield[0] >> relative & 1]), b'\x01')

        self.assertEqual(address_state.ots_bitfield_used_page, 1)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_update_used_page_in_address_state2(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

        alice_xmss = get_alice_xmss(12)
        address = alice_xmss.address
        address_state = OptimizedAddressState.get_default(address)
        addresses_state = {address: address_state}
        paginated_bitfield = PaginatedBitfield(True, self.state._db)
        paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
        self.assertEqual(address_state.ots_bitfield_used_page, 0)

        factor = min(config.dev.ots_tracking_per_page, 2 ** alice_xmss.height)
        for i in range(0, 2 ** alice_xmss.height):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, (i + 1) // factor)

        self.assertEqual(address_state.ots_bitfield_used_page, 4)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_update_used_page_in_address_state3(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

        alice_xmss = get_alice_xmss(12)
        address = alice_xmss.address
        address_state = OptimizedAddressState.get_default(address)
        addresses_state = {address: address_state}
        paginated_bitfield = PaginatedBitfield(True, self.state._db)
        paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
        self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(3072, 2 ** alice_xmss.height):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(2048, 3072):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(1024, 2048):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(0, 1024):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            if i + 1 == 1024:
                self.assertEqual(address_state.ots_bitfield_used_page, 4)

        self.assertEqual(address_state.ots_bitfield_used_page, 4)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_update_used_page_in_address_state4(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

        alice_xmss = get_alice_xmss(12)
        address = alice_xmss.address
        address_state = OptimizedAddressState.get_default(address)
        addresses_state = {address: address_state}
        paginated_bitfield = PaginatedBitfield(True, self.state._db)
        paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
        self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(2048, 3072):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(1024, 2048):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

        for i in range(0, 1024):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            if i + 1 == 1024:
                self.assertEqual(address_state.ots_bitfield_used_page, 3)

        for i in range(3072, 2 ** alice_xmss.height):
            paginated_bitfield.set_ots_key(addresses_state, address, i)
            if i + 1 == 2 ** alice_xmss.height:
                self.assertEqual(address_state.ots_bitfield_used_page, 4)
            else:
                self.assertEqual(address_state.ots_bitfield_used_page, 3)

        self.assertEqual(address_state.ots_bitfield_used_page, 4)



from unittest import TestCase

from mock import Mock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.MessageTransaction import MessageTransaction
from tests.misc.helper import get_alice_xmss, set_qrl_dir

logger.initialize_default()


class TestMessageTransactionStateChanges(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100
        self.addresses_state = {
            self.alice.address: alice_address_state
        }

        self.params = {
            "message_hash": b'Test Message',
            "addr_to": None,
            "fee": 1,
            "xmss_pk": self.alice.pk
        }
        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def test_apply_message_txn(self):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

    def test_revert_message_txn(self):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)

        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

    def test_validate_tx(self):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from qrl.generated import qrl_pb2 as qrl__pb2


class PublicAPIStub(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Public API used by clients (wallet/cli/etc)
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetNodeState = channel.unary_unary(
        '/qrl.PublicAPI/GetNodeState',
        request_serializer=qrl__pb2.GetNodeStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetNodeStateResp.FromString,
        )
    self.GetKnownPeers = channel.unary_unary(
        '/qrl.PublicAPI/GetKnownPeers',
        request_serializer=qrl__pb2.GetKnownPeersReq.SerializeToString,
        response_deserializer=qrl__pb2.GetKnownPeersResp.FromString,
        )
    self.GetPeersStat = channel.unary_unary(
        '/qrl.PublicAPI/GetPeersStat',
        request_serializer=qrl__pb2.GetPeersStatReq.SerializeToString,
        response_deserializer=qrl__pb2.GetPeersStatResp.FromString,
        )
    self.GetStats = channel.unary_unary(
        '/qrl.PublicAPI/GetStats',
        request_serializer=qrl__pb2.GetStatsReq.SerializeToString,
        response_deserializer=qrl__pb2.GetStatsResp.FromString,
        )
    self.GetAddressState = channel.unary_unary(
        '/qrl.PublicAPI/GetAddressState',
        request_serializer=qrl__pb2.GetAddressStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetAddressStateResp.FromString,
        )
    self.GetOptimizedAddressState = channel.unary_unary(
        '/qrl.PublicAPI/GetOptimizedAddressState',
        request_serializer=qrl__pb2.GetAddressStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetOptimizedAddressStateResp.FromString,
        )
    self.GetMultiSigAddressState = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigAddressState',
        request_serializer=qrl__pb2.GetMultiSigAddressStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetMultiSigAddressStateResp.FromString,
        )
    self.IsSlave = channel.unary_unary(
        '/qrl.PublicAPI/IsSlave',
        request_serializer=qrl__pb2.IsSlaveReq.SerializeToString,
        response_deserializer=qrl__pb2.IsSlaveResp.FromString,
        )
    self.GetObject = channel.unary_unary(
        '/qrl.PublicAPI/GetObject',
        request_serializer=qrl__pb2.GetObjectReq.SerializeToString,
        response_deserializer=qrl__pb2.GetObjectResp.FromString,
        )
    self.GetLatestData = channel.unary_unary(
        '/qrl.PublicAPI/GetLatestData',
        request_serializer=qrl__pb2.GetLatestDataReq.SerializeToString,
        response_deserializer=qrl__pb2.GetLatestDataResp.FromString,
        )
    self.PushTransaction = channel.unary_unary(
        '/qrl.PublicAPI/PushTransaction',
        request_serializer=qrl__pb2.PushTransactionReq.SerializeToString,
        response_deserializer=qrl__pb2.PushTransactionResp.FromString,
        )
    self.TransferCoins = channel.unary_unary(
        '/qrl.PublicAPI/TransferCoins',
        request_serializer=qrl__pb2.TransferCoinsReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.ParseAddress = channel.unary_unary(
        '/qrl.PublicAPI/ParseAddress',
        request_serializer=qrl__pb2.ParseAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.ParseAddressResp.FromString,
        )
    self.GetChainStats = channel.unary_unary(
        '/qrl.PublicAPI/GetChainStats',
        request_serializer=qrl__pb2.GetChainStatsReq.SerializeToString,
        response_deserializer=qrl__pb2.GetChainStatsResp.FromString,
        )
    self.GetAddressFromPK = channel.unary_unary(
        '/qrl.PublicAPI/GetAddressFromPK',
        request_serializer=qrl__pb2.GetAddressFromPKReq.SerializeToString,
        response_deserializer=qrl__pb2.GetAddressFromPKResp.FromString,
        )
    self.GetMultiSigCreateTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigCreateTxn',
        request_serializer=qrl__pb2.MultiSigCreateTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetMultiSigSpendTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigSpendTxn',
        request_serializer=qrl__pb2.MultiSigSpendTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetMultiSigVoteTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigVoteTxn',
        request_serializer=qrl__pb2.MultiSigVoteTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetMessageTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetMessageTxn',
        request_serializer=qrl__pb2.MessageTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTokenTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetTokenTxn',
        request_serializer=qrl__pb2.TokenTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTransferTokenTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetTransferTokenTxn',
        request_serializer=qrl__pb2.TransferTokenTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetSlaveTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetSlaveTxn',
        request_serializer=qrl__pb2.SlaveTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetLatticeTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetLatticeTxn',
        request_serializer=qrl__pb2.LatticeTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTransaction = channel.unary_unary(
        '/qrl.PublicAPI/GetTransaction',
        request_serializer=qrl__pb2.GetTransactionReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTransactionResp.FromString,
        )
    self.GetMiniTransactionsByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetMiniTransactionsByAddress',
        request_serializer=qrl__pb2.GetMiniTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetMiniTransactionsByAddressResp.FromString,
        )
    self.GetTransactionsByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetTransactionsByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTransactionsByAddressResp.FromString,
        )
    self.GetTokensByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetTokensByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTokensByAddressResp.FromString,
        )
    self.GetSlavesByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetSlavesByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetSlavesByAddressResp.FromString,
        )
    self.GetLatticePKsByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetLatticePKsByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetLatticePKsByAddressResp.FromString,
        )
    self.GetMultiSigAddressesByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigAddressesByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetMultiSigAddressesByAddressResp.FromString,
        )
    self.GetMultiSigSpendTxsByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetMultiSigSpendTxsByAddress',
        request_serializer=qrl__pb2.GetMultiSigSpendTxsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetMultiSigSpendTxsByAddressResp.FromString,
        )
    self.GetVoteStats = channel.unary_unary(
        '/qrl.PublicAPI/GetVoteStats',
        request_serializer=qrl__pb2.GetVoteStatsReq.SerializeToString,
        response_deserializer=qrl__pb2.GetVoteStatsResp.FromString,
        )
    self.GetInboxMessagesByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetInboxMessagesByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetInboxMessagesByAddressResp.FromString,
        )
    self.GetBalance = channel.unary_unary(
        '/qrl.PublicAPI/GetBalance',
        request_serializer=qrl__pb2.GetBalanceReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBalanceResp.FromString,
        )
    self.GetTotalBalance = channel.unary_unary(
        '/qrl.PublicAPI/GetTotalBalance',
        request_serializer=qrl__pb2.GetTotalBalanceReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTotalBalanceResp.FromString,
        )
    self.GetOTS = channel.unary_unary(
        '/qrl.PublicAPI/GetOTS',
        request_serializer=qrl__pb2.GetOTSReq.SerializeToString,
        response_deserializer=qrl__pb2.GetOTSResp.FromString,
        )
    self.GetHeight = channel.unary_unary(
        '/qrl.PublicAPI/GetHeight',
        request_serializer=qrl__pb2.GetHeightReq.SerializeToString,
        response_deserializer=qrl__pb2.GetHeightResp.FromString,
        )
    self.GetBlock = channel.unary_unary(
        '/qrl.PublicAPI/GetBlock',
        request_serializer=qrl__pb2.GetBlockReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBlockResp.FromString,
        )
    self.GetBlockByNumber = channel.unary_unary(
        '/qrl.PublicAPI/GetBlockByNumber',
        request_serializer=qrl__pb2.GetBlockByNumberReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBlockByNumberResp.FromString,
        )


class PublicAPIServicer(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Public API used by clients (wallet/cli/etc)
  """

  def GetNodeState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetKnownPeers(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPeersStat(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetStats(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetOptimizedAddressState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigAddressState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def IsSlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetObject(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLatestData(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TransferCoins(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ParseAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetChainStats(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressFromPK(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigCreateTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigSpendTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigVoteTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMessageTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransferTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetSlaveTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLatticeTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMiniTransactionsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTokensByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetSlavesByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLatticePKsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigAddressesByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMultiSigSpendTxsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetVoteStats(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetInboxMessagesByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBalance(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTotalBalance(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetOTS(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetHeight(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlock(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockByNumber(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PublicAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetNodeState': grpc.unary_unary_rpc_method_handler(
          servicer.GetNodeState,
          request_deserializer=qrl__pb2.GetNodeStateReq.FromString,
          response_serializer=qrl__pb2.GetNodeStateResp.SerializeToString,
      ),
      'GetKnownPeers': grpc.unary_unary_rpc_method_handler(
          servicer.GetKnownPeers,
          request_deserializer=qrl__pb2.GetKnownPeersReq.FromString,
          response_serializer=qrl__pb2.GetKnownPeersResp.SerializeToString,
      ),
      'GetPeersStat': grpc.unary_unary_rpc_method_handler(
          servicer.GetPeersStat,
          request_deserializer=qrl__pb2.GetPeersStatReq.FromString,
          response_serializer=qrl__pb2.GetPeersStatResp.SerializeToString,
      ),
      'GetStats': grpc.unary_unary_rpc_method_handler(
          servicer.GetStats,
          request_deserializer=qrl__pb2.GetStatsReq.FromString,
          response_serializer=qrl__pb2.GetStatsResp.SerializeToString,
      ),
      'GetAddressState': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressState,
          request_deserializer=qrl__pb2.GetAddressStateReq.FromString,
          response_serializer=qrl__pb2.GetAddressStateResp.SerializeToString,
      ),
      'GetOptimizedAddressState': grpc.unary_unary_rpc_method_handler(
          servicer.GetOptimizedAddressState,
          request_deserializer=qrl__pb2.GetAddressStateReq.FromString,
          response_serializer=qrl__pb2.GetOptimizedAddressStateResp.SerializeToString,
      ),
      'GetMultiSigAddressState': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigAddressState,
          request_deserializer=qrl__pb2.GetMultiSigAddressStateReq.FromString,
          response_serializer=qrl__pb2.GetMultiSigAddressStateResp.SerializeToString,
      ),
      'IsSlave': grpc.unary_unary_rpc_method_handler(
          servicer.IsSlave,
          request_deserializer=qrl__pb2.IsSlaveReq.FromString,
          response_serializer=qrl__pb2.IsSlaveResp.SerializeToString,
      ),
      'GetObject': grpc.unary_unary_rpc_method_handler(
          servicer.GetObject,
          request_deserializer=qrl__pb2.GetObjectReq.FromString,
          response_serializer=qrl__pb2.GetObjectResp.SerializeToString,
      ),
      'GetLatestData': grpc.unary_unary_rpc_method_handler(
          servicer.GetLatestData,
          request_deserializer=qrl__pb2.GetLatestDataReq.FromString,
          response_serializer=qrl__pb2.GetLatestDataResp.SerializeToString,
      ),
      'PushTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.PushTransaction,
          request_deserializer=qrl__pb2.PushTransactionReq.FromString,
          response_serializer=qrl__pb2.PushTransactionResp.SerializeToString,
      ),
      'TransferCoins': grpc.unary_unary_rpc_method_handler(
          servicer.TransferCoins,
          request_deserializer=qrl__pb2.TransferCoinsReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'ParseAddress': grpc.unary_unary_rpc_method_handler(
          servicer.ParseAddress,
          request_deserializer=qrl__pb2.ParseAddressReq.FromString,
          response_serializer=qrl__pb2.ParseAddressResp.SerializeToString,
      ),
      'GetChainStats': grpc.unary_unary_rpc_method_handler(
          servicer.GetChainStats,
          request_deserializer=qrl__pb2.GetChainStatsReq.FromString,
          response_serializer=qrl__pb2.GetChainStatsResp.SerializeToString,
      ),
      'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressFromPK,
          request_deserializer=qrl__pb2.GetAddressFromPKReq.FromString,
          response_serializer=qrl__pb2.GetAddressFromPKResp.SerializeToString,
      ),
      'GetMultiSigCreateTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigCreateTxn,
          request_deserializer=qrl__pb2.MultiSigCreateTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetMultiSigSpendTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigSpendTxn,
          request_deserializer=qrl__pb2.MultiSigSpendTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetMultiSigVoteTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigVoteTxn,
          request_deserializer=qrl__pb2.MultiSigVoteTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetMessageTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetMessageTxn,
          request_deserializer=qrl__pb2.MessageTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetTokenTxn,
          request_deserializer=qrl__pb2.TokenTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransferTokenTxn,
          request_deserializer=qrl__pb2.TransferTokenTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetSlaveTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetSlaveTxn,
          request_deserializer=qrl__pb2.SlaveTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetLatticeTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetLatticeTxn,
          request_deserializer=qrl__pb2.LatticeTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransaction,
          request_deserializer=qrl__pb2.GetTransactionReq.FromString,
          response_serializer=qrl__pb2.GetTransactionResp.SerializeToString,
      ),
      'GetMiniTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetMiniTransactionsByAddress,
          request_deserializer=qrl__pb2.GetMiniTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetMiniTransactionsByAddressResp.SerializeToString,
      ),
      'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionsByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetTransactionsByAddressResp.SerializeToString,
      ),
      'GetTokensByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetTokensByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetTokensByAddressResp.SerializeToString,
      ),
      'GetSlavesByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetSlavesByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetSlavesByAddressResp.SerializeToString,
      ),
      'GetLatticePKsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetLatticePKsByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetLatticePKsByAddressResp.SerializeToString,
      ),
      'GetMultiSigAddressesByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigAddressesByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetMultiSigAddressesByAddressResp.SerializeToString,
      ),
      'GetMultiSigSpendTxsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetMultiSigSpendTxsByAddress,
          request_deserializer=qrl__pb2.GetMultiSigSpendTxsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetMultiSigSpendTxsByAddressResp.SerializeToString,
      ),
      'GetVoteStats': grpc.unary_unary_rpc_method_handler(
          servicer.GetVoteStats,
          request_deserializer=qrl__pb2.GetVoteStatsReq.FromString,
          response_serializer=qrl__pb2.GetVoteStatsResp.SerializeToString,
      ),
      'GetInboxMessagesByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetInboxMessagesByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetInboxMessagesByAddressResp.SerializeToString,
      ),
      'GetBalance': grpc.unary_unary_rpc_method_handler(
          servicer.GetBalance,
          request_deserializer=qrl__pb2.GetBalanceReq.FromString,
          response_serializer=qrl__pb2.GetBalanceResp.SerializeToString,
      ),
      'GetTotalBalance': grpc.unary_unary_rpc_method_handler(
          servicer.GetTotalBalance,
          request_deserializer=qrl__pb2.GetTotalBalanceReq.FromString,
          response_serializer=qrl__pb2.GetTotalBalanceResp.SerializeToString,
      ),
      'GetOTS': grpc.unary_unary_rpc_method_handler(
          servicer.GetOTS,
          request_deserializer=qrl__pb2.GetOTSReq.FromString,
          response_serializer=qrl__pb2.GetOTSResp.SerializeToString,
      ),
      'GetHeight': grpc.unary_unary_rpc_method_handler(
          servicer.GetHeight,
          request_deserializer=qrl__pb2.GetHeightReq.FromString,
          response_serializer=qrl__pb2.GetHeightResp.SerializeToString,
      ),
      'GetBlock': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlock,
          request_deserializer=qrl__pb2.GetBlockReq.FromString,
          response_serializer=qrl__pb2.GetBlockResp.SerializeToString,
      ),
      'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockByNumber,
          request_deserializer=qrl__pb2.GetBlockByNumberReq.FromString,
          response_serializer=qrl__pb2.GetBlockByNumberResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.PublicAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class AdminAPIStub(object):
  """This is a place holder for testing/instrumentation APIs
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """


class AdminAPIServicer(object):
  """This is a place holder for testing/instrumentation APIs
  """


def add_AdminAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.AdminAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from unittest import TestCase

import simplejson as json

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.misc import logger
from qrl.core.VoteStats import VoteStats
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.MultiSigAddressState import MultiSigAddressState
from qrl.core.txs.multisig.MultiSigSpend import MultiSigSpend
from qrl.core.txs.multisig.MultiSigVote import MultiSigVote
from tests.core.txs.testdata import test_json_MultiSigVote
from qrl.generated.qrl_pb2 import SlaveMetadata
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_qrl_dir, set_hard_fork_block_number

logger.initialize_default()


class TestMultiSigVote(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestMultiSigVote, self).__init__(*args, **kwargs)
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()
        self.random = get_alice_xmss(4)
        self.random_signer = get_bob_xmss(4)
        self.signatories = [self.alice.address, self.bob.address, self.random.address]
        self.weights = [20, 30, 10]
        self.threshold = 30

    def test_create(self):
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=0,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)
        self.assertIsInstance(tx, MultiSigVote)

    def test_to_json(self):
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=0,
                                 xmss_pk=self.alice.pk)
        txjson = tx.to_json()
        self.assertEqual(json.loads(test_json_MultiSigVote), json.loads(txjson))

    def test_validate_custom(self):
        """
        MultiSigCreate _validate_custom() only checks if fee == 0
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=0,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)
        result = tx._validate_custom()
        self.assertTrue(result)

    @set_hard_fork_block_number()
    def test_validate_extended(self):
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=None,
                                         batch=None)

        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        tx._data.multi_sig_vote.unvote = True
        result = tx._validate_extended(state_container)
        self.assertFalse(result)

        tx._data.multi_sig_vote.unvote = False
        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        alice_address_state.pbdata.balance = 0
        result = tx._validate_extended(state_container)
        self.assertFalse(result)

        alice_address_state.pbdata.balance = 5
        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        state_container.block_number = 15000
        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        state_container.block_number = 15001
        result = tx._validate_extended(state_container)
        self.assertFalse(result)

    @set_hard_fork_block_number()
    def test_validate_all(self):
        """
        Test for Validate Extended when transaction has been signed by slave.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.random.pk,
                                 master_addr=self.alice.address)
        tx.sign(self.random)
        tx._data.nonce = 1
        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        random_address_state = OptimizedAddressState.get_default(address=self.random.address)
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            self.random.address: random_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        slaves = Indexer(b'slave', None)
        slaves.data[(self.alice.address, self.random.pk)] = SlaveMetadata(access_type=0)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=slaves,
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=False,
                                         my_db=self.state._db,
                                         batch=None)

        result = tx.validate_all(state_container)
        self.assertTrue(result)

        tx._data.nonce = 2
        result = tx.validate_all(state_container)
        self.assertFalse(result)  # False as nonce is invalid

    def test_apply(self):
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address,
                                                                                            tx.ots_key))

        tx.apply(self.state, state_container)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address,
                                                                                           tx.ots_key))
        self.assertIn(spend_tx.txhash, state_container.votes_stats)
        vote_stats = state_container.votes_stats[spend_tx.txhash]
        unvote, index = vote_stats.get_unvote_by_address(tx.addr_from)
        self.assertNotEqual(index, -1)
        self.assertFalse(unvote)
        self.assertEqual(vote_stats.shared_key, spend_tx.txhash)
        self.assertEqual(vote_stats.total_weight, 4)
        self.assertEqual(vote_stats.signatories, multi_sig_address_state.signatories)

    def test_revert(self):
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address,
                                                                                            tx.ots_key))

        tx.apply(self.state, state_container)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address,
                                                                                           tx.ots_key))
        self.assertIn(spend_tx.txhash, state_container.votes_stats)
        vote_stats = state_container.votes_stats[spend_tx.txhash]
        unvote, index = vote_stats.get_unvote_by_address(tx.addr_from)
        self.assertNotEqual(index, -1)
        self.assertFalse(unvote)
        self.assertEqual(vote_stats.shared_key, spend_tx.txhash)
        self.assertEqual(vote_stats.total_weight, 4)
        self.assertEqual(vote_stats.signatories, multi_sig_address_state.signatories)

        tx.revert(self.state, state_container)
        self.assertIn(spend_tx.txhash, state_container.votes_stats)
        vote_stats = state_container.votes_stats[spend_tx.txhash]
        unvote, index = vote_stats.get_unvote_by_address(tx.addr_from)
        self.assertNotEqual(index, -1)
        self.assertTrue(unvote)
        self.assertEqual(vote_stats.shared_key, spend_tx.txhash)
        self.assertEqual(vote_stats.total_weight, 0)
        self.assertEqual(vote_stats.signatories, multi_sig_address_state.signatories)

        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address,
                                                                                            tx.ots_key))

    def test_affected_address(self):
        # This transaction can only involve 2 addresses.
        affected_addresses = set()
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)
        tx.set_affected_address(affected_addresses)

        self.assertEqual(1, len(affected_addresses))
        self.assertIn(self.alice.address, affected_addresses)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

# leveldb code for maintaining account state data
import plyvel
import os

from qrl.core import config
from qrl.core.misc import logger

__author__ = 'pete'


class DB:
    def __init__(self, db_dir=None):
        self.db_dir = os.path.join(config.user.data_dir, config.dev.db_name)
        if db_dir:
            self.db_dir = db_dir
        logger.info('DB path: %s', self.db_dir)

        os.makedirs(self.db_dir, exist_ok=True)

        try:
            self.db = plyvel.DB(self.db_dir, max_open_files=1000, lru_cache_size=5 * 1024)
        except Exception:
            self.db = plyvel.DB(self.db_dir,
                                max_open_files=1000,
                                lru_cache_size=5 * 1024,
                                create_if_missing=True,
                                compression='snappy')
            self.db.put(b'state_version', str(1).encode())

    def close(self):
        del self.db

    def open(self, db_dir=None):
        if db_dir:
            self.db_dir = db_dir
        self.db = plyvel.DB(self.db_dir, max_open_files=1000, lru_cache_size=5 * 1024)

    def RangeIter(self, key_obj_start, key_obj_end):
        if not isinstance(key_obj_start, bytes):
            key_obj_start = key_obj_start.encode()

        if not isinstance(key_obj_end, bytes):
            key_obj_end = key_obj_end.encode()

        return self.db.RangeIter(key_obj_start, key_obj_end)

    def get_db_keys(self, include_value: bool):
        return self.db.RangeIter(include_value=include_value)

    def delete(self, key_obj: bytes, batch=None):
        if batch:
            batch.delete(key_obj)
        else:
            self.db.delete(key_obj)

    def put_raw(self, key, value, batch=None):
        if batch:
            batch.put(key, value)
        else:
            self.db.put(key, value)

    def get_raw(self, key):
        if isinstance(key, str):
            key = bytes(key, 'utf-8')
        value = self.db.get(key)
        if value is None:
            raise KeyError
        return value

    def get_batch(self):
        return self.db.write_batch()

    @staticmethod
    def write_batch(batch, sync=True):
        batch.write()



from collections import namedtuple

from qrl.generated import qrl_pb2
from qrl.core.misc import logger
from qrl.core.State import State


class ProposalRecord:
    def __init__(self, protobuf_block=None):
        self._data = protobuf_block
        if protobuf_block is None:
            self._data = qrl_pb2.ProposalRecord()

        counter_mapping = namedtuple("counter_mapping", ["get", "update"])

        self._counter_by_name = {
            b'p_proposal_tx_hash': counter_mapping(self.number_of_tx_hashes,
                                                   self.update_number_of_tx_hashes),
        }

    def pbdata(self):
        return self._data

    def get_counter_by_name(self, name: bytes):
        return self._counter_by_name[name].get()

    def update_counter_by_name(self, name, value=1, subtract=False):
        self._counter_by_name[name].update(value, subtract)

    def number_of_tx_hashes(self):
        return self._data.number_of_tx_hashes

    def update_number_of_tx_hashes(self, value=1, subtract=False):
        if subtract:
            self._data.number_of_tx_hashes -= value
        else:
            self._data.number_of_tx_hashes += value

    def serialize(self):
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrl_pb2.ProposalRecord()
        pbdata.ParseFromString(bytes(data))
        return ProposalRecord(pbdata)

    @staticmethod
    def put_state(state: State, key, proposal_record, batch):
        try:
            state._db.put_raw(key, proposal_record.serialize(), batch)
        except Exception as e:
            raise Exception("[ProposalRecord] Exception in put_state %s", e)

    @staticmethod
    def get_state(state: State, key: bytes):
        try:
            data = state._db.get_raw(key)
            return ProposalRecord.deserialize(data)
        except KeyError:
            logger.debug('[get_state] ProposalRecord not found')
        except Exception as e:
            logger.error('[get_state] %s', e)

        return ProposalRecord()

    @staticmethod
    def get_key(block_number, activation_delay):
        """
            activation_block_number is the block number after which config will be activated
            so an activation block number 10 means the config will be activated only after adding
            block number 10 into the block chain
        """
        activation_block_number = block_number + activation_delay
        return b'proposal_record_block_number_' + activation_block_number.to_bytes(8, byteorder='big', signed=False)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os
from unittest import TestCase

from mock import patch

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.qrlnode import QRLNode
from qrl.generated.qrlbase_pb2 import GetNodeInfoReq
from qrl.services.BaseService import BaseService
from tests.misc.helper import replacement_getTime

logger.initialize_default()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestBaseAPI(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBaseAPI, self).__init__(*args, **kwargs)

    def test_getNodeInfo(self):
        qrlnode = QRLNode(mining_address=b'')

        service = BaseService(qrlnode)
        response = service.GetNodeInfo(request=GetNodeInfoReq, context=None)

        self.assertEqual(config.dev.version, response.version)

        proto_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                  os.path.pardir,
                                                  os.path.pardir,
                                                  "src", "qrl", "protos", "qrl.proto"))

        with open(proto_path, 'r') as content_file:
            proto_content = content_file.read()

        self.assertEqual(proto_content, response.grpcProto)



from pyqryptonight.pyqryptonight import PoWHelper


class DifficultyTracker(object):
    def __init__(self):
        pass

    @staticmethod
    def get_target(current_difficulty, dev_config):
        ph = PoWHelper(kp=dev_config.kp,
                       set_point=dev_config.block_timing_in_seconds)
        return ph.getTarget(current_difficulty)

    @staticmethod
    def get(measurement, parent_difficulty, dev_config):
        ph = PoWHelper(kp=dev_config.kp,
                       set_point=dev_config.block_timing_in_seconds)

        current_difficulty = ph.getDifficulty(measurement=measurement,
                                              parent_difficulty=parent_difficulty)

        current_target = ph.getTarget(current_difficulty)
        return current_difficulty, current_target



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qrl.core import config
from qrl.core.misc import logger
from qrl.generated import qrl_pb2


class PaginatedData:
    def __init__(self, name: bytes, write_access: bool, db):
        self.name = name
        self.key_value = dict()
        self.db = db
        self.write_access = write_access

    def reset_key_value(self):
        self.key_value = dict()

    def get_value(self, key: bytes, count: int) -> list:
        storage_key = self.generate_key(key, count)
        return self.key_value[storage_key]

    def insert(self, address_state, value: bytes):
        key = address_state.address
        count = address_state.get_counter_by_name(self.name)
        storage_key = self.generate_key(address_state.address, count)

        if storage_key not in self.key_value:
            self.key_value[storage_key] = self.get_paginated_data(key, count)

        self.key_value[storage_key].append(value)

        address_state.update_counter_by_name(self.name)

    def remove(self, address_state, value: bytes):
        address_state.update_counter_by_name(self.name, value=1, subtract=True)

        key = address_state.address
        count = address_state.get_counter_by_name(self.name)
        storage_key = self.generate_key(address_state.address, count)

        if storage_key not in self.key_value:
            self.key_value[storage_key] = self.get_paginated_data(key, count)

        if self.key_value[storage_key][-1] != value:
            logger.warning("Expected value %s", self.key_value[storage_key][-1])
            logger.warning("Found value %s", value)
            raise Exception("Unexpected value into storage")

        del self.key_value[storage_key][-1]

    def put_paginated_data(self, batch) -> bool:
        key_value = self.key_value
        self.key_value = dict()
        for storage_key in key_value:
            value = key_value[storage_key]
            if len(value) == 0:
                self.delete(storage_key, batch)
            self.put(storage_key, value, batch)

        return True

    def get_paginated_data(self, key, count) -> list:
        storage_key = self.generate_key(key, count)
        try:
            pbData = self.db.get_raw(storage_key)
            data_list = qrl_pb2.DataList()
            data_list.ParseFromString(bytes(pbData))
            return list(data_list.values)
        except KeyError:
            return []
        except Exception as e:
            logger.error('[get_paginated_data] Exception for %s', self.name)
            logger.exception(e)
            raise

    def generate_key(self, key, count: int):
        page = count // config.dev.data_per_page
        return self.name + b'_' + key + b'_' + str(page).encode()

    def put(self, storage_key, value, batch):
        if not self.write_access:
            return
        data_list = qrl_pb2.DataList(values=value)
        self.db.put_raw(storage_key,
                        data_list.SerializeToString(),
                        batch)

    def delete(self, storage_key, batch):
        if not self.write_access:
            return
        self.db.delete(storage_key,
                       batch)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from pyqrllib.pyqrllib import hstr2bin
from unittest import TestCase

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.MultiSigAddressState import MultiSigAddressState
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.txs.multisig.MultiSigCreate import MultiSigCreate
from qrl.core.State import State
from tests.misc.helper import set_qrl_dir, get_bob_xmss, get_alice_xmss, get_random_xmss

logger.initialize_default()


class TestMultiSigAddressState(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

    def test_generate_multi_sig_address(self):
        creation_tx_hash = bytes(hstr2bin("5a4c37ef7e5b7cc5a2a58ab730269ed8"
                                          "f4cbf08a005dc3508e31465535e1d6bb"))
        address = MultiSigAddressState.generate_multi_sig_address(creation_tx_hash)
        expected_address = bytes(hstr2bin("1100003674370317e1cac0ca13f896ab5b6472a"
                                          "261ba0d2b2961d3adba1b9060f6e8f7fe2088fb"))
        self.assertEqual(address, expected_address)
        self.assertFalse(OptimizedAddressState.address_is_valid(address))

    def test_address_is_valid(self):
        address = bytes(hstr2bin("110000000000000000000000000000000000000"
                                 "000000000000000000000000000000000000000"))
        self.assertFalse(OptimizedAddressState.address_is_valid(address))

    def test_get_multi_sig_address_state_by_address(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()
        random_xmss = get_random_xmss()

        signatories = [alice_xmss.address, bob_xmss.address]
        weights = [20, 20]
        threshold = 21
        multi_sig_tx = MultiSigCreate.create(signatories,
                                             weights,
                                             threshold,
                                             0,
                                             random_xmss.pk)
        multi_sig_tx.sign(random_xmss)
        multi_sig_address_state = MultiSigAddressState.get_default(multi_sig_tx.txhash,
                                                                   signatories,
                                                                   weights,
                                                                   threshold)
        AddressState.put_address_state(self.state, multi_sig_address_state)
        multi_sig_address_state2 = MultiSigAddressState.get_multi_sig_address_state_by_address(
            self.state._db,
            MultiSigAddressState.generate_multi_sig_address(multi_sig_tx.txhash))

        self.assertEqual(multi_sig_address_state.pbdata, multi_sig_address_state2.pbdata)

    def test_put_multi_sig_addresses_state(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()
        random_xmss = get_random_xmss()

        signatories = [alice_xmss.address, bob_xmss.address]
        weights = [20, 20]
        threshold = 21
        multi_sig_tx = MultiSigCreate.create(signatories,
                                             weights,
                                             threshold,
                                             0,
                                             random_xmss.pk)
        multi_sig_tx.sign(random_xmss)
        multi_sig_address_state = MultiSigAddressState.get_default(multi_sig_tx.txhash,
                                                                   signatories,
                                                                   weights,
                                                                   threshold)

        multi_sig_addresses_state = {multi_sig_address_state.address: multi_sig_address_state}
        AddressState.put_addresses_state(self.state, multi_sig_addresses_state)

        multi_sig_address_state2 = MultiSigAddressState.get_multi_sig_address_state_by_address(
            self.state._db,
            MultiSigAddressState.generate_multi_sig_address(multi_sig_tx.txhash))
        self.assertEqual(multi_sig_address_state.pbdata, multi_sig_address_state2.pbdata)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import copy

from pyqryptonight.pyqryptonight import Qryptominer, SOLUTION

from qrl.core.config import DevConfig, user as user_config
from qrl.core.Block import Block
from qrl.core.misc import logger


class CNv1Miner(Qryptominer):
    def __init__(self,
                 pre_block_logic,
                 mining_address: bytes,
                 mining_thread_count,
                 lock):
        super().__init__()

        self.pre_block_logic = pre_block_logic  # FIXME: Circular dependency with node.py

        self._mining_address = mining_address
        self._mining_thread_count = mining_thread_count

        self._mining_block = None
        self.setForcedSleep(user_config.mining_pause)
        self._dev_config = None
        self.lock = lock

    def start_mining(self,
                     mining_block: Block,
                     current_target: bytes,
                     dev_config: DevConfig):
        try:
            logger.debug('start_mining - TRY LOCK')
            with self.lock:
                logger.debug('start_mining - LOCKED')
                self.cancel()

                mining_blob = mining_block.mining_blob(dev_config)
                nonce_offset = mining_block.mining_nonce_offset(dev_config)

                self._dev_config = dev_config
                self._mining_block = mining_block
                work_seq_id = self.start(input=mining_blob,
                                         nonceOffset=nonce_offset,
                                         target=current_target,
                                         thread_count=self._mining_thread_count)

                logger.debug("MINING START [{}]".format(work_seq_id))

        except Exception as e:
            logger.warning("Exception in start_mining")
            logger.exception(e)

        logger.debug('start_mining - UNLOCKED')

    def handleEvent(self, event):
        # NOTE: This function usually runs in the context of a C++ thread
        if event.type == SOLUTION:
            logger.debug('handleEvent - TRY LOCK')
            if not self.lock.acquire(blocking=False):
                logger.debug('handleEvent - SKIP')
                return False

            try:
                logger.debug('handleEvent - LOCKED')

                logger.debug('Solution Found %s', event.nonce)
                logger.info('Hash Rate: %s H/s', self.hashRate())
                cloned_block = copy.deepcopy(self._mining_block)
                cloned_block.set_nonces(self._dev_config, event.nonce, 0)
                logger.debug("Blob           %s", cloned_block)
                logger.info('Block #%s nonce: %s', cloned_block.block_number, event.nonce)
                self.pre_block_logic(cloned_block)
            except Exception as e:
                logger.warning("Exception in solutionEvent")
                logger.exception(e)
            finally:
                logger.debug('handleEvent - UNLOCK')
                self.lock.release()

        return True
